/**********************************************************************
 * imap4_out.c                                               March 1999
 * Horms                                             horms@vergenet.net
 *
 * Functions to communicate with upstream IMAP4 server
 *
 * Released under the terms of the GNU GPL
 *
 **********************************************************************/

#include "imap4_out.h"

/**********************************************************************
 * imap4_authenticate
 * Authenticate user with backend imap4 server
 * pre: in_fd: File descriptor to read server responses from
 *      out_fd: File descriptor to write to server on
 *      err_fd: File descriptor to write error to
 *      tag:    ignored 
 *      pw:     structure with username and passwd
 * post: 1: on success
 *       0: on failure
 *       -1 on error
 **********************************************************************/

int imap4_out_authenticate(
  const int in_fd, 
  const int out_fd, 
  const struct passwd *pw,
  protocol_t *protocol
){
  token_t *ok;
  queue *q;
  char *read_string;
  char *greeting_string;
  int status;

  read_string=NULL;
  greeting_string=NULL;
  status=-1;

  if((ok=create_token(ok))==NULL){
    syslog(LOG_WARNING, "imap4_out_authenticate: create_token");
    goto leave;
  }
  assign_token(ok, (PERDITION_USTRING) IMAP4_OK, strlen(IMAP4_OK), -1);

  if((status=imap4_out_response(in_fd, IMAP4_UNTAGED, ok, &q))<0){
    syslog(LOG_WARNING, "imap4_out_authenticate: imap4_out_response");
    goto leave;
  }
  else if(!status){
    status=0;
    goto leave;
  }

  if((read_string=queue_to_string(q, read_string))==NULL){
    syslog(LOG_WARNING, "imap4_out_authenticate: queue_to_string");
    goto leave;
  }
 
  if((greeting_string=greeting_str(greeting_string, protocol))==NULL){
    syslog(LOG_WARNING, "imap4_out_authenticate: greeting_str");
    goto leave;
  }

  if((status=strcmp(read_string, greeting_string))==0){
    syslog(LOG_WARNING, "Loop detected, abandoning connection");
    goto leave;
  }

  syslog(
    LOG_INFO,
    "read_string=\"%s\", greeting_string=\"%s\", status=%d",
    read_string,
    greeting_string,
    status
  );

  if(write_str(
    out_fd, 
    5, 
    IMAP4_ONE_TIME_TAG, 
    " LOGIN ",
    pw->pw_name,
    " ",
    pw->pw_passwd)<0
  ){
    syslog(LOG_WARNING, "imap4_out_authenticate: imap4_write");
    status=-1;
    goto leave;
  }

  if((status=imap4_out_response(in_fd, IMAP4_ONE_TIME_TAG, ok, &q))<0){
    syslog(LOG_WARNING, "imap4_out_authenticate: imap4_out_response 2");
  }
  destroy_queue(q, DESTROY_TOKEN);

  leave:
  str_free(&read_string);
  str_free(&greeting_string);
  destroy_token(&ok);
  return(status);
}
  

/**********************************************************************
 * imap4_out_response
 * Compare a respnse from a server with the desired response
 * pre: in_fd: file descriptor to read from
 *      tag_string: tag expected from server
 *      desired_token: token expected from server
 *      q: resulting queue is stored here
 * post: 1 : tag and desired token found
 *       0: tag and desired token not found
 *       -1: on error
 **********************************************************************/

int imap4_out_response(
  const int in_fd, 
  const char *tag_string,
  const token_t *desired_token,
  queue **q
){
  int status;
  int tagged;
  token_t *t;
  char *server_tag_string;

  *q=NULL;
  server_tag_string=NULL;
  t=NULL;
  status=0;

  /*
   * tagged set to 0 if an untagged message is expected
   * set to non-zero otherwise
   */
  tagged=strcmp(tag_string, IMAP4_UNTAGED);

  /*Check tag*/
  while(1){
    if((*q=read_line(in_fd, *q))==NULL){
      syslog(LOG_WARNING, "imap4_out_response: read_line");
      return(-1);
    }
  
    if((*q=pop(*q, (queue_t *)&t))==NULL){
      syslog(LOG_WARNING, "imap4_out_response: pop");
      return(-1);
    }

    if((server_tag_string=token_to_string(t, server_tag_string))==NULL){
      syslog(LOG_WARNING, "imap4_out_response: token_to_string");
      goto leave;
    }
    unassign_token(t);
    destroy_token(&t);
  
    if(tagged && !strcmp(server_tag_string, IMAP4_UNTAGED)){
      destroy_queue(*q, DESTROY_TOKEN);
      continue;
    }
  
    if(strcmp(server_tag_string, tag_string)){
      syslog(LOG_WARNING, "imap4_out_resonse: invalid tag from server");
      goto leave;
    }
  
    break;
  }

  if((*q=pop(*q, (queue_t *)&t))==NULL){
    syslog(LOG_WARNING, "imap4_out_response: pop");
    return(-1);
  }
  
  /*Check token*/
  status=token_cmp(desired_token, t);

  leave:
  destroy_token(&t);
  if(status!=1){
    destroy_queue(*q, DESTROY_TOKEN);
    *q=NULL;
  }
  str_free(&server_tag_string);
  return(status);
}
