/**********************************************************************
 * token.c                                               September 1999
 * Horms                                             horms@vergenet.net
 *
 * Token to encapsulate a byte string
 *
 * Licenced under the GNU GPL
 *
 **********************************************************************/

#include "token.h"



/**********************************************************************
 * create_token
 * create an empty token
 *
 * 8 bit clean
 **********************************************************************/

token_t *create_token(token_t *t){
  if((t=(token_t *)malloc(sizeof(token_t)))==NULL){
    syslog(LOG_WARNING, "create_token: malloc");
    return(NULL);
  }
  t->n=0;
  t->buf=NULL;
  return(t);
}


/**********************************************************************
 * assign_token
 * place bytes into a token
 *
 * 8 bit clean
 **********************************************************************/

void assign_token(
  token_t *t, 
  unsigned char * buf, 
  const ssize_t n,
  const int eol
){
  t->n=n;
  t->eol=eol;
  if(n==0){
    t->buf=NULL;
  }
  else {
    t->buf=buf;
  }
}


/**********************************************************************
 * unassign_token
 * make a token empty
 *
 * useful if you want to destroy a token but not what it contains
 *
 * 8 bit clean
 **********************************************************************/

void unassign_token(token_t *t){
  assign_token(t, (unsigned char *)NULL, (ssize_t)0, 0);
}


/**********************************************************************
 * destroy_token
 * if the token is null no nothing
 * if buf is non-NULL then free it
 * free the token
 *
 * 8 bit clean
 **********************************************************************/

void destroy_token(token_t **t){
  if(*t==NULL){
    return;
  }

  if((*t)->buf!=NULL){
    free((*t)->buf);
  }

  free(*t);
  *t=NULL;
}


/**********************************************************************
 * write_token
 * write a token to fd
 *
 * Pre: fd: File descriptor to write to
 *      token: token to write
 * Return: -1 on error
 *         0 otherwise
 *
 * 8 bit clean
 **********************************************************************/

int write_token(const int fd, const token_t *t){
  if(write_strn(fd, t->buf, t->n)){
    syslog(LOG_WARNING, "write_token: write_strn");
    return(-1);
  }
  
  return(0);
}


/**********************************************************************
 * read_token
 * read a token in from fd
 * ' ' will terminate a token
 * '\r' is ignored
 * '\"' is ignored
 * '\n' will terminate a token and set the eol element to 1
 * all other characters are considered to be a part of the token
 *
 * 8 bit clean
 **********************************************************************/

token_t *read_token(const int fd, token_t *t){
  unsigned char *buffer;
  int bytes_read;
  int offset;
  int eol;
  fd_set read_template;
  fd_set except_template;
  struct timeval timeout;
  int status;

  extern int idle_timeout;

  if(
    (buffer=(unsigned char *)malloc(sizeof(unsigned char)*BUFFER_SIZE))==NULL
  ){
    syslog(
      LOG_WARNING, 
      "read_token: malloc buffer: %s", 
      strerror(errno)
    );
    return(NULL);
  }

  offset=0;
  eol=0;
  while(offset<BUFFER_SIZE){
    FD_ZERO(&read_template);
    FD_SET(fd, &read_template);

    FD_ZERO(&except_template);
    FD_SET(fd, &except_template);

    timeout.tv_sec=idle_timeout;
    timeout.tv_usec=0;

    status=select(
      FD_SETSIZE,
      &read_template,
      NULL,
      &except_template,
      idle_timeout?&timeout:NULL
    );
    if(status<0){
      if(errno!=EINTR){
        syslog(LOG_WARNING, "read_token: select: %s, Exiting", strerror(errno));
        daemon_exit_cleanly(-1);
      }
      goto leave;
    }
    else if(FD_ISSET(fd, &except_template)){
      syslog(LOG_WARNING, "read_token: error on file descriptor");
      goto leave;
    }
    else if(status==0){
      syslog(LOG_WARNING, "read_token: idle timeout");
      daemon_exit_cleanly(0);
    }

    /*If we get this far fd must be ready for reading*/
    if((bytes_read=read(fd, buffer+offset, 1))<0){
      syslog(
        LOG_WARNING, 
        "read_token: error reading input",
        strerror(errno)
      );
      goto leave;
    }
    if(bytes_read==0){
      syslog(
        LOG_WARNING,
        "read_token: zero bytes read",
        strerror(errno)
      );
      goto leave;
    }

    if(*(buffer+offset)=='\n'){
      eol=1;
      break;
    }
    else if(*(buffer+offset)=='\r'){
      continue;
    }
    else if(*(buffer+offset)=='\"'){
      continue;
    }
    else if(*(buffer+offset)==' '){
      break;
    }
    else{
      offset++;
    }
  }

  if((t=create_token(t))==NULL){
    syslog(LOG_WARNING, "read_token: create_token");
    goto leave;
  }

  assign_token(t, buffer, offset, eol);
    
  return(t);

  /*Escape hatch if something goes wrong*/
  leave:
  free(buffer);
  return(NULL);
}


/**********************************************************************
 * token_is_eol
 * Check if the token is at the end of a line
 * return 1 if the token is at the end of a linr
 *        0 otherwise
 **********************************************************************/

int token_is_eol(const token_t *t){
  return(t->eol);
}


/**********************************************************************
 * token_cmp
 * compare two tokens
 * return 1 is they are the same, 0 otherwise
 * eol field will be ignored if eiter token has eol 
 * set to -1 (don't care)
 *
 * Not 8 bit clean as it is case insensitive using toupper
 **********************************************************************/

int token_cmp(const token_t *a, const token_t *b){
  int i;

  if(a->n!=b->n){
    return(0);
  }

  if((a->eol!=b->eol) && (a->eol>-1) && (b->eol>-1)){
    return(0);
  }

  for(i=0;i<a->n;i++){
    if(toupper(*(a->buf+i))!=toupper(*(b->buf+i))){
      return(0);
    }
  }

  return(1);
}


/**********************************************************************
 * token_is_null
 * test if a token is null
 * return 0 if token is null
 *        -1 otherwise
 * eol field is ignored
 *
 * 8 bit clean
 **********************************************************************/

int token_is_null(const token_t *t){
  return(t->n==0?0:-1);
}


/**********************************************************************
 * token_to_string
 * dump the buffer in a token into a \0 terminated string
 * string will be dynamically allocated
 *
 * Not 8 bit clean
 **********************************************************************/

char *token_to_string(const token_t *t, char *string){
  if((string=strn_to_str(string, (char *)t->buf, t->n))==NULL){
    syslog(LOG_WARNING, "token_to_string: strn_to_str");
    return(NULL);
  }
  return(string);
}

